#!/usr/bin/env python

#
# Copyright 2012 Frank Adelstein.  All Rights Reserved.
#
# Permission to use, copy, modify, and distribute this
# software is hereby granted without fee, provided that
# the copyright notice and permission notice are not removed.
#
# Last changed: 2013-11-11 by FNA
#   updated FSM table
#

import pygtk
pygtk.require('2.0')
import gtk

import string
import subprocess

import signal, os
import time

import sys

import bananaconst
import bananamenu
import bananatile
import bananabag

import bananahistory
import bananaevent
import bananafsm

FSM = (('start', 'click', ('on_tile', 'this_tile_selected'), None, 'clicked'),
       ('start', 
        'click', ('on_tile', 'not_this_tile_selected'), 
        ('select_tile', 'redraw',), 'clickedselect'),
       ('start', 'click', ('not_on_tile',), None, 'emptyclick'),

       ('clicked', 'release', (), (), 'start'),
       ('clicked', 'move', 
        None, ('save_position', 'update_position',), 'move'),

       ('clickedselect', 
        'release', ('on_tile',), (), 'start'),
       ('clickedselect', 'move', 
        None, ('save_position', 'update_position',), 'move'),

       ('emptyclick', 
        'release', ('one_selected',), 
        ('move_tile', 'set_move_hist', 'check_spelling'), 'start'),
       ('emptyclick', 
        'release', ('none_selected',), None, 'start'),
       ('emptyclick', 
        'release', ('plural_selected',), ('deselect_all',), 'start'),
       ('emptyclick', 
        'move', ('one_selected',), ('start_rubberband',), 'rubberband'),
       ('emptyclick', 
        'move', ('any_selected',), ('start_rubberband',), 'rubberband'),
       ('emptyclick', 
        'move', ('none_selected',), ('start_rubberband',), 'rubberband'),

       ('move', 'move', None, ('update_position',), 'move'),
       ('move', 
        'release', None, 
        ('update_position', 'fix_position', 'hist_spell_if_moved'), 'start'),

       ('rubberband', 'move', None, ('rubberband',), 'rubberband'),
       ('rubberband', 
        'release', ('one_selected',), ('end_rubberband_special', 'redraw'), 
        'start'),
       ('rubberband', 
        'release', None, ('end_rubberband', 'redraw'), 'start'),
      )

class Banana:

    # instance variables
    pangolayout = 0
    button1 = 0
    rubberbanding = 0
    startdragx = -1
    startdragy = -1
    lastdragx = -1
    lastdragy = -1
    totaltiles = 0
    tilesout = 0
    selected = []
    tiles = 0

    spellmode = 1
    allselected = 0

    controlkey = 0
    shiftkey = 0

    tilewidth = 30

    grid = 1
    timer = 0
    history = 0
    debug = False

    # methods

    # quit method
    def destroy(self, widget):
        gtk.main_quit()

    ###########################
    # reset all the tiles and #
    # read them in again      #
    # extra argument so it    #
    # can be invoked by       #
    # do_dialog()             #
    ###########################
    def reset(self, optional=None):
        
        self.tilesout = 0
        self.tiles = []
        self.totaltiles = 0
        self.tilebag = []
        self.selected = []
        self.history = bananahistory.History()
        self.fsm.reset()

        # read in the letter config file and return a list of
        # the letters to select in order
        bananabag.Tilebag(os.path.join(os.path.dirname(sys.argv[0]), bananaconst.tilefile), self);
        
        self.labels['tilelabel'].set_text("Tiles out: " + str(self.tilesout) + " / Tiles total: " + str(self.totaltiles))
        
        self.redraw_cb()
        return

    def unselect_all(self):
        for tile in self.tiles:
            tile.selected = 0
        self.selected = []
        self.redraw_cb()
        self.allselected = 0
        return

    # handler for key release events (shift and control)
    def key_release_handler(self, w, event):
        keyname = gtk.gdk.keyval_name(event.keyval)
#        print "Key %s (%d) was released" % (keyname, event.keyval)
        if keyname == "Shift_L" or keyname == "Shift_R":
            self.shiftkey = 0
        if keyname == "Control_L" or keyname == "Control_R":
            self.controlkey = 0

    # handler for keys, especially escape/quit/space
    def key_press_handler(self, w, event):
        keyname = gtk.gdk.keyval_name(event.keyval)
        # print "Key %s (%d) was pressed" % (keyname, event.keyval)
        if keyname == 'Escape':
            quit_dialog(self)

        if keyname == 'space':
            # draw a new tile from the bag (add a Tile object)
            if len(self.tilebag):
                # unselect any existing tiles
                self.unselect_all()
                letter = self.tilebag.pop(0)
                self.tiles.append(bananatile.Tile(letter, self.labels['area'], 50, 50, self.tilewidth, self.tilewidth))
                # make drawn tile be the selected tile
                self.tiles[-1].selected = 1
                self.selected = [self.tiles[-1]]
                self.redraw_cb()
                self.tilesout += 1
                self.labels['tilelabel'].set_text("Tiles out: " + str(self.tilesout) + " / Tiles total: " + str(self.totaltiles))

            else:
                print "no more letters"

        if keyname == 'd':
            # deal out initial set of tiles if none are out
            if not len(self.tiles):
                self.deal()

        # toggle debug mode
        if keyname == 'D':
            if self.debug:
                print "debug mode off"
                self.debug = False
            else:
                print "debug mode on"
                self.debug = True

        # print out undo queue
        if keyname == "q" and self.debug:
            print self.history.list()

        ###print tile DEBUG information ####
        if keyname == 'p' and self.debug:
            print "letter(id), top, bottom, left, right"
            for x in self.tiles:
                print "%s (%s), %s, %s, %s, %s" % (x.letter, x, x.top, x.bottom, x.left, x.right)

        ###print selected tile DEBUG information ####
        if keyname == 's' and self.debug:
            print "selected tiles"
            for x in self.selected:
                print "%s" % (x.letter)

        # move selected tiles left, right, up, or down 
        if keyname == "Left":
            if len(self.selected):
                eventlist = []
                # move selected tiles left one square
                for tile in self.selected:
                    x1 = tile.x
                    y1 = tile.y
                    tile.x -= self.tilewidth + 1
                    event = bananaevent.Move(tile, x1, y1, tile.x, tile.y)
                    eventlist.append(event)
                # save in do/undo list
                self.history.do(eventlist)
                # redraw so the color changes take effect
                self.redraw_cb()

        if keyname == "Right":
            if len(self.selected):
                eventlist = []
                # move selected tiles left one square
                for tile in self.selected:
                    x1 = tile.x
                    y1 = tile.y
                    tile.x += self.tilewidth + 1
                    event = bananaevent.Move(tile, x1, y1, tile.x, tile.y)
                    eventlist.append(event)
                # save in do/undo list
                self.history.do(eventlist)
                # redraw so the color changes take effect
                self.redraw_cb()

        if keyname == "Up":
            if len(self.selected):
                eventlist = []
                # move selected tiles up one square 
                for tile in self.selected:
                    x1 = tile.x
                    y1 = tile.y
                    tile.y -= self.tilewidth + 1
                    event = bananaevent.Move(tile, x1, y1, tile.x, tile.y)
                    eventlist.append(event)
                # save in do/undo list
                self.history.do(eventlist)
                # redraw so the color changes take effect
                self.redraw_cb()

        if keyname == "Down":
            if len(self.selected):
                eventlist = []
                # move selected tiles down one square
                for tile in self.selected:
                    x1 = tile.x
                    y1 = tile.y
                    tile.y += self.tilewidth + 1
                    event = bananaevent.Move(tile, x1, y1, tile.x, tile.y)
                    eventlist.append(event)
                # save in do/undo list
                self.history.do(eventlist)
                # redraw so the color changes take effect
                self.redraw_cb()


        if keyname == "Shift_L" or keyname == "Shift_R":
            self.shiftkey = 1
        if keyname == "Control_L" or keyname == "Control_R":
            self.controlkey = 1

    ################
    # start method #
    ################ 
    def __init__(self):

        # set up variables
        self.labels = dict()
        self.tiles = []
        self.tilebag = []
        self.history = bananahistory.History()

        self.fsm = bananafsm.FSM(FSM, 'start')

        # read in the letter config file and return a list of
        # the letters to select in order
        bananabag.Tilebag(os.path.join(os.path.dirname(sys.argv[0]), bananaconst.tilefile), self);


        # create a main window
        window = gtk.Window(gtk.WINDOW_TOPLEVEL)
        window.set_title(bananaconst.name + "" + bananaconst.version)
        window.set_border_width(1)
        #window.set_size_request(bananaconst.width,bananaconst.height)
        window.connect("destroy", self.destroy)
        self.labels['window'] = window

        # set the window's icon
        window.set_icon_from_file(os.path.join(os.path.dirname(sys.argv[0]), bananaconst.iconfile))

        window.connect('key_press_event', self.key_press_handler)
        window.connect('key_release_event', self.key_release_handler)
        
        # tooltips can be reused for all tips
        tooltips = gtk.Tooltips()
        self.labels['tooltips'] = tooltips

        # main vbox
        vbox = gtk.VBox(False, 0)
        window.add(vbox)
        vbox.modify_bg(gtk.STATE_NORMAL, gtk.gdk.Color(green=1.0))

        accel_group = gtk.AccelGroup()
        window.add_accel_group(accel_group)

        # make a menu
        menu = bananamenu.Menu(vbox, self, accel_group)

        area = gtk.DrawingArea()
        self.labels['area'] = area
        area.set_size_request(bananaconst.width,bananaconst.height)
        self.pangolayout = area.create_pango_layout("")
        # note that this widget CAN expand and get filled when resized
        vbox.pack_start(area, True, True, 0)

        area.set_events(gtk.gdk.BUTTON1_MOTION_MASK | 
                        gtk.gdk.BUTTON_PRESS_MASK |
                        gtk.gdk.BUTTON_RELEASE_MASK)

        area.connect("expose-event", self.area_expose_cb)

        ############################################
        # Callback functions are defined within    #
        # the scope of __init__() because of goofy #
        # scoping rules and limits of passing      #
        # things into the callback.                #
        ############################################

        # handle button press/release events
        def button_notify(widget, event):
            if event.type == gtk.gdk.BUTTON_PRESS and event.button == 1:
                event_name = "click"
            elif event.type == gtk.gdk.BUTTON_RELEASE and event.button == 1:
                event_name = "release"
            else:
                # skip other button events (besides button 1 press/release)
                return
            # pass event to finite state machine to handle
            self.fsm.transition(self, event_name, event)
            return

        # handle mouse motion events
        def motion_notify(widget, event):
            # pass event to finite state machine to handle
            self.fsm.transition(self, "move", event)
            return

        ###########################
        # widget has changed size #
        ###########################
        def configure_notify(widget, event):
            # print "widget is now (x,y, width, height) (%d, %d, %d, %d)" % (event.x, event.y, event.width, event.height)
#            area.set_size_request(event.width,event.height - 20)
            bananaconst.width = event.width
            bananaconst.height = event.height
            return False;

        ##################################
        # continuing with __init__() ... #
        ##################################
        area.connect_object("motion_notify_event", motion_notify, self)
        area.connect_object("button_release_event", button_notify, self)
        area.connect_object("button_press_event",  button_notify, self)
        area.connect_object("configure_event",  configure_notify, self)

        # set it to a green background
        area.modify_bg(gtk.STATE_NORMAL, gtk.gdk.color_parse('#00db00'))

        # hbox within vbox to store stats (clock, number of tiles
        hbox = gtk.HBox(False, 0)
        vbox.pack_end(hbox, False, False, 0)
        hbox.set_size_request(-1, 20)
        self.labels['hbox'] = hbox

        label = gtk.Label("Tiles out: " + str(self.tilesout) + " / Tiles total: " + str(self.totaltiles))
        hbox.pack_start(label, False, False, 0)
        self.labels['tilelabel'] = label

        label = gtk.Label("tick tock")
        hbox.pack_end(label, False, False, 0)
        self.labels['timelabel'] = label


        window.show_all()

        self.makeGCs()



        #################################
        # handler for alarm-clock timer #
        #################################
        def alarmHandler(signum, frame):
            # increment the timer (by 1 second)
            self.timer += 1
            # format the elapsed time string
            timetext = '%02d:%02d:%02d' % (int(self.timer/3600), 
                                           int((self.timer /60) % 60), 
                                           self.timer % 60)
            # upate the time label widget
            self.labels['timelabel'].set_text(timetext)
            # set the timer to go off in 1 second
            signal.alarm(1)

            return


        # set alarm signal handler
        signal.signal(signal.SIGALRM, alarmHandler)
        signal.alarm(1)

        ### end __init__() ###

    ###########################################
    # method to redraw the drawing area,      #
    # used when a tile is dragged or making   #
    # a selection rubberband.                 #
    ###########################################
    def redraw_cb(self):
        area = self.labels['area']

        # activate backing-store for the updates to prevent display flicker
        rect = gtk.gdk.Rectangle(0,0,area.get_allocation().width, area.get_allocation().height)
        area.window.begin_paint_rect(rect)

        area.window.clear()

        if self.grid:
            # make vertical lines
            step = self.tilewidth + 1
            for col in range(step, bananaconst.width, step):
                area.window.draw_line(self.whitegc, col, 0, col, bananaconst.height)
            # make horizontal lines (keep same step, tiles are square)
            for row in range(step, bananaconst.height, step):
                area.window.draw_line(self.whitegc, 0, row, bananaconst.width, row)

        for tile in self.tiles:
            tile.draw(self)

        # end backing-store anti-flicker buffering
        area.window.end_paint()

    ################################
    # callback for exposure events #
    # Of limited use.              #
    ################################
    def area_expose_cb(self, area, event):
        self.redraw_cb();
        return

        self.style = area.get_style()
        self.gc = self.style.fg_gc[gtk.STATE_NORMAL]
        self.gc.set_values(line_width=5, fill=gtk.gdk.SOLID)
        self.xorgc = area.window.new_gc(function = gtk.gdk.XOR)
        gdk_color = self.xorgc.get_colormap().alloc_color('white')
        self.xorgc.set_foreground(gdk_color)
#        gdk_color = self.xorgc.get_colormap().alloc_color('white')
#        self.gc.set_foreground(gdk_color)
        area.window.draw_rectangle(self.gc, False, 200, 200, 40, 20)
        self.pangolayout.set_text("rectangle")
        area.window.draw_layout(self.gc, 205,  280, self.pangolayout)
        return

    #################################
    # function to deal out 15 tiles #
    # at the start of a game        #
    #################################
    def deal(self):
        for i in range(3):
            for j in range(5):
                if len(self.tilebag):
                    letter = self.tilebag.pop(0)
                    self.tiles.append(bananatile.Tile(letter, self.labels['area'], 50 + j*50, 300 + i*50, self.tilewidth, self.tilewidth))
                    self.redraw_cb()
                    self.tilesout += 1
                    self.labels['tilelabel'].set_text("Tiles out: " + str(self.tilesout) + " / Tiles total: " + str(self.totaltiles))
                    # Process all pending events to force GTK to draw tiles
                    while gtk.events_pending():
                        gtk.main_iteration(False)
                    time.sleep(0.3)
                else:
                    print "no more letters"
            
        return

    #################################
    # Create graphic contexts (GCs) #
    # that are used in drawing      #
    #################################
    def makeGCs(self):
        area = self.labels['area']
        style = area.get_style()

        # standard black GC
        self.gc = style.fg_gc[gtk.STATE_NORMAL]
        self.gc.set_values(line_width=5, fill=gtk.gdk.SOLID)
        self.pangolayout = area.create_pango_layout("")

        # create a white GC
        self.whitegc = area.window.new_gc()
        gdk_color = self.whitegc.get_colormap().alloc_color('antique white')
        self.whitegc.set_foreground(gdk_color)
        
        # create a gold GC (to use for selected tiles)
        self.selectedgc = area.window.new_gc()
        gdk_color = self.selectedgc.get_colormap().alloc_color('gold')
        self.selectedgc.set_foreground(gdk_color)

        # create a red GC (to use for misspelled tiles)
        self.redgc = area.window.new_gc()
        gdk_color = self.redgc.get_colormap().alloc_color('red')
        self.redgc.set_foreground(gdk_color)

        # create XOR Graphics Context
        self.xorgc = area.window.new_gc(function = gtk.gdk.XOR)
        gdk_color = self.xorgc.get_colormap().alloc_color('white')
        self.xorgc.set_foreground(gdk_color)

        return

    ###########################
    ### end of Banana class ###
    ###########################

# pop up a confirm to quit dialog
def quit_dialog(obj):

    do_dialog(obj, "Really quit?", "Really quit the program?", obj.destroy)
    return
    
###############
# pop up a dialog to confirm something, then take the action if 'yes' selected
###############
def do_dialog(obj, string1, string2, action):

    dialog = gtk.Dialog(string1, None, 
                        gtk.DIALOG_DESTROY_WITH_PARENT | gtk.DIALOG_MODAL,
                        (gtk.STOCK_NO, gtk.RESPONSE_NO, 
                         gtk.STOCK_YES, gtk.RESPONSE_YES))

    labeltxt = gtk.Label(string2)
    dialog.vbox.pack_start(labeltxt, True, True, 0)

    dialog.show_all()

    response = dialog.run()
    dialog.destroy()

    # quit if they selected "yes"
    if response == gtk.RESPONSE_YES:
        action(None)

    return


# allow GUI moves only to places that don't overlap other tiles (end drag)
# 'dump' action puts 1 tile back in bag, takes 3 more
# clock updates continuously (pause/unpause allowed)




def main():

    # fix change where icons are NOT displayed for some reason
    settings = gtk.settings_get_default()
    settings.props.gtk_button_images = True

    # reopen stdout file descriptor with write mode
    # and 0 as the buffer size (unbuffered)
    sys.stdout = os.fdopen(sys.stdout.fileno(), 'w', 0) 

    gtk.main()
    return 0

if __name__ == "__main__":
    Banana()
    main()

